#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis {
  [StorableClass]
  public abstract class DataAnalysisProblemData : ParameterizedNamedItem,
    IDataAnalysisProblemData {
    private const string DatasetParameterName = "Dataset";
    private const string InputVariablesParameterName = "Input variables";
    private const string TrainingSamplesStartParameterName = "Training partition start";
    private const string TrainingSamplesEndParameterName = "Training partition end";
    private const string TestSamplesStartParameterName = "Test partition start";
    private const string TestSamplesEndParameterName = "Test partition end";

    #region parameter properties
    public IValueParameter<Dataset> DatasetParameter {
      get { return (IValueParameter<Dataset>)Parameters[DatasetParameterName]; }
    }
    public IValueParameter<ICheckedItemList<StringValue>> InputVariablesParameter {
      get { return (IValueParameter<ICheckedItemList<StringValue>>)Parameters[InputVariablesParameterName]; }
    }
    public IValueParameter<IntValue> TrainingSamplesStartParameter {
      get { return (IValueParameter<IntValue>)Parameters[TrainingSamplesStartParameterName]; }
    }
    public IValueParameter<IntValue> TrainingSamplesEndParameter {
      get { return (IValueParameter<IntValue>)Parameters[TrainingSamplesEndParameterName]; }
    }
    public IValueParameter<IntValue> TestSamplesStartParameter {
      get { return (IValueParameter<IntValue>)Parameters[TestSamplesStartParameterName]; }
    }
    public IValueParameter<IntValue> TestSamplesEndParameter {
      get { return (IValueParameter<IntValue>)Parameters[TestSamplesEndParameterName]; }
    }
    #endregion

    #region propeties
    public Dataset Dataset {
      get { return DatasetParameter.Value; }
    }
    public IEnumerable<string> AllowedInputVariables {
      get { return InputVariablesParameter.Value.CheckedItems.Select(i => i.Value.Value); }
    }

    public int TrainingSamplesStart {
      get { return TrainingSamplesStartParameter.Value.Value; }
      set { TrainingSamplesStartParameter.Value.Value = value; }
    }
    public int TrainingSamplesEnd {
      get { return TrainingSamplesEndParameter.Value.Value; }
      set { TrainingSamplesEndParameter.Value.Value = value; }
    }

    public int TestSamplesStart {
      get { return TestSamplesStartParameter.Value.Value; }
      set { TestSamplesStartParameter.Value.Value = value; }
    }
    public int TestSamplesEnd {
      get { return TestSamplesEndParameter.Value.Value; }
      set { TestSamplesEndParameter.Value.Value = value; }
    }

    public IEnumerable<int> TrainingIndizes {
      get {
        return Enumerable.Range(TrainingSamplesStart, TrainingSamplesEnd - TrainingSamplesStart)
                         .Where(i => i >= 0 && i < Dataset.Rows && (i < TestSamplesStart || TestSamplesEnd <= i));
      }
    }
    public IEnumerable<int> TestIndizes {
      get {
        return Enumerable.Range(TestSamplesStart, TestSamplesEnd - TestSamplesStart)
           .Where(i => i >= 0 && i < Dataset.Rows);
      }
    }
    #endregion

    protected DataAnalysisProblemData(DataAnalysisProblemData original, Cloner cloner)
      : base(original, cloner) {
      RegisterEventHandlers();
    }
    [StorableConstructor]
    protected DataAnalysisProblemData(bool deserializing) : base(deserializing) { }
    protected DataAnalysisProblemData(Dataset dataset, IEnumerable<string> allowedInputVariables) {
      if (allowedInputVariables.Except(dataset.VariableNames).Any())
        throw new ArgumentException("All allowed input variables must be present in the dataset.");

      ICheckedItemList<StringValue> inputVariableList = new CheckedItemList<StringValue>(dataset.VariableNames.Select(x => new StringValue(x))).AsReadOnly();
      foreach (StringValue inputVariable in inputVariableList) {
        inputVariableList.SetItemCheckedState(inputVariable, allowedInputVariables.Contains(inputVariable.Value));
      }

      Parameters.Add(new ValueParameter<Dataset>(DatasetParameterName, dataset));
      Parameters.Add(new ValueParameter<ICheckedItemList<StringValue>>(InputVariablesParameterName, inputVariableList));
      Parameters.Add(new ValueParameter<IntValue>(TrainingSamplesStartParameterName, new IntValue(0)));
      Parameters.Add(new ValueParameter<IntValue>(TrainingSamplesEndParameterName, new IntValue(dataset.Rows)));
      Parameters.Add(new ValueParameter<IntValue>(TestSamplesStartParameterName, new IntValue(dataset.Rows)));
      Parameters.Add(new ValueParameter<IntValue>(TestSamplesEndParameterName, new IntValue(dataset.Rows)));

      RegisterEventHandlers();
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterEventHandlers();
    }

    #region changed event propagation
    private void RegisterEventHandlers() {
      DatasetParameter.ValueChanged += new EventHandler(Dataset_Changed);
      TrainingSamplesStartParameter.ValueChanged += new EventHandler(PartitionParameter_ValueChanged);
      TrainingSamplesEndParameter.ValueChanged += new EventHandler(PartitionParameter_ValueChanged);
      TestSamplesStartParameter.ValueChanged += new EventHandler(PartitionParameter_ValueChanged);
      TestSamplesEndParameter.ValueChanged += new EventHandler(PartitionParameter_ValueChanged);

      InputVariablesParameter.ValueChanged += new EventHandler(InputVariablesParameter_ValueChanged);
      RegisterInputVariablesParameterValueEventHandlers();

      TrainingSamplesStartParameter.Value.ValueChanged += new EventHandler(Partitions_ValueChanged);
      TrainingSamplesEndParameter.Value.ValueChanged += new EventHandler(Partitions_ValueChanged);
      TestSamplesStartParameter.Value.ValueChanged += new EventHandler(Partitions_ValueChanged);
      TestSamplesEndParameter.Value.ValueChanged += new EventHandler(Partitions_ValueChanged);
    }
    private void RegisterInputVariablesParameterValueEventHandlers() {
      InputVariablesParameter.Value.CheckedItemsChanged += new CollectionItemsChangedEventHandler<IndexedItem<StringValue>>(InputVariables_Changed);
      InputVariablesParameter.Value.CollectionReset += new CollectionItemsChangedEventHandler<IndexedItem<StringValue>>(InputVariables_Changed);
      InputVariablesParameter.Value.ItemsAdded += new CollectionItemsChangedEventHandler<IndexedItem<StringValue>>(InputVariables_Changed);
      InputVariablesParameter.Value.ItemsMoved += new CollectionItemsChangedEventHandler<IndexedItem<StringValue>>(InputVariables_Changed);
      InputVariablesParameter.Value.ItemsRemoved += new CollectionItemsChangedEventHandler<IndexedItem<StringValue>>(InputVariables_Changed);
      InputVariablesParameter.Value.ItemsReplaced += new CollectionItemsChangedEventHandler<IndexedItem<StringValue>>(InputVariables_Changed);
    }

    private void PartitionParameter_ValueChanged(object sender, EventArgs e) {
      IntValue value = (IntValue)sender;
      value.ValueChanged += new EventHandler(Partitions_ValueChanged);
      OnChanged();
    }
    private void InputVariablesParameter_ValueChanged(object sender, EventArgs e) {
      RegisterInputVariablesParameterValueEventHandlers();
    }

    private void Partitions_ValueChanged(object sender, EventArgs e) {
      OnChanged();
    }
    private void InputVariables_Changed(object sender, CollectionItemsChangedEventArgs<IndexedItem<StringValue>> e) {
      OnChanged();
    }
    private void Dataset_Changed(object sender, EventArgs e) {
      OnChanged();
    }

    public event EventHandler Changed;
    protected virtual void OnChanged() {
      var listeners = Changed;
      if (listeners != null) listeners(this, EventArgs.Empty);
    }
    #endregion
  }
}
